# -*- coding: utf-8 -*-
import os

import pytest

from lektor.builder import Builder
from lektor.context import Context
from lektor.project import Project


@pytest.fixture
def lektor_env():
    site_path = os.path.join(os.path.dirname(__file__), 'test-site')
    return Project.from_path(site_path).make_env(load_plugins=False)


@pytest.fixture
def lektor_pad(lektor_env):
    return lektor_env.new_pad()


@pytest.fixture
def lektor_builder(lektor_pad, tmp_path):
    return Builder(lektor_pad, str(tmp_path))


@pytest.fixture
def lektor_build_state(lektor_builder):
    with lektor_builder.new_build_state() as build_state:
        yield build_state


@pytest.fixture
def lektor_source(lektor_pad):
    return lektor_pad.root


@pytest.fixture
def lektor_artifact(lektor_build_state, lektor_source):
    return lektor_build_state.new_artifact(
        'test.html', source_obj=lektor_source)


@pytest.fixture
def lektor_context(lektor_artifact):
    with Context(lektor_artifact) as ctx:
        yield ctx



# -*- coding: utf-8 -*-
import re

import pytest

from inifile import IniFile
from lektor.markdown import (
    ImprovedRenderer,
    MarkdownConfig,
    )

from lektor_responsive_image import (
    ResponsiveImage,
    resolve_image,
    ResponsiveImageMixin,
    ResponsiveImagePlugin,
    )


class DummyImage(object):
    def __init__(self, width, height, format='jpeg', quality=None):
        self.width = width
        self.height = height
        self.format = format
        self.quality = quality

    def thumbnail(self, width, quality=None, upscale=None):
        height = int(self.height * width / self.width + 0.5)
        return DummyImage(width, height, format=self.format, quality=quality)


class DummyContext(object):
    def url_to(self, image):
        return 'img-{}.jpg'.format(image.width)


class TestResponsiveImage(object):
    @pytest.fixture
    def width(self):
        return 2400

    @pytest.fixture
    def image(self, width):
        height = int(3 * width / 4)
        return DummyImage(width, height)

    @pytest.fixture
    def sizes(self):
        return None

    @pytest.fixture
    def resp_image(self, image, sizes):
        config = ResponsiveImage.DEFAULT_CONFIG.copy()
        config['sizes'] = sizes
        return ResponsiveImage(image, config, ctx=DummyContext())

    def test_init_no_config(self, image):
        resp_image = ResponsiveImage(image)
        assert resp_image.config == ResponsiveImage.DEFAULT_CONFIG
        assert resp_image.config is not ResponsiveImage.DEFAULT_CONFIG

    @pytest.mark.parametrize('width', [480])
    def test_attrs_no_srcset(self, resp_image):
        assert 'srcset' not in resp_image.attrs

    @pytest.mark.parametrize('width', [800])
    def test_attrs_srcset(self, resp_image):
        srcset = resp_image.attrs['srcset']
        assert srcset == 'img-480.jpg 480w, img-800.jpg 800w'

    @pytest.mark.parametrize('sizes', [None])
    def test_attrs_no_sizes(self, resp_image):
        assert 'sizes' not in resp_image.attrs

    @pytest.mark.parametrize('sizes', ['600px'])
    def test_attrs_sizes(self, resp_image, sizes):
        resp_image.attrs['sizes'] == sizes

    @pytest.mark.parametrize('sizes', ['50vw', None])
    def test_sizes(self, resp_image, sizes):
        assert resp_image.sizes == sizes

    @pytest.mark.parametrize('width, widths', [
        (120, [120]),
        (480, [480]),
        (1024, [480, 800, 1024]),
        (3600, [480, 800, 1200, 2400]),
        ])
    def test_iter_widths(self, resp_image, widths):
        assert list(resp_image.iter_widths()) == widths

    def test_default_image_attrs(self, resp_image):
        assert resp_image.default_image_attrs() == {
            'width': 1200,
            'height': 900,
            'src': 'img-1200.jpg',
            }

    def test_resize_image(self, resp_image, image):
        assert resp_image.resize_image(image.width) is image
        assert resp_image.resize_image(image.width + 1) is image
        resized = resp_image.resize_image(image.width - 1)
        assert resized.width == image.width - 1
        assert resized.quality == 92


class Test_resolve_image(object):
    def test_no_record(self):
        assert resolve_image(None, 'test.jpg') is None

    def test_external_url(self, lektor_pad):
        assert resolve_image(lektor_pad.root,
                             'http://example.com/test.jpg') is None

    def test_not_image(self, lektor_pad):
        assert resolve_image(lektor_pad.root, 'dummy.pdf') is None

    def test_resolve(self, lektor_pad):
        image = resolve_image(lektor_pad.root, "test.jpg")
        assert image.path == '/test.jpg'
        assert image.width == 800

    def test_resolve_absolute(self, lektor_pad):
        about = lektor_pad.get('/about')
        image = resolve_image(about, "/test.jpg")
        assert image.path == '/test.jpg'

    def test_resolve_relative(self, lektor_pad):
        about = lektor_pad.get('/about')
        image = resolve_image(about, "../test.jpg")
        assert image.path == '/test.jpg'


@pytest.fixture
def load_plugin(lektor_env):
    lektor_env.plugin_controller.instanciate_plugin(
        'responsive-image', ResponsiveImagePlugin)


class TestResponsiveImageMixin(object):
    @pytest.fixture
    def record(self, lektor_pad):
        return lektor_pad.root

    @pytest.fixture
    def mixin(self, record):
        class Renderer(ResponsiveImageMixin, ImprovedRenderer):
            def __init__(self, record):
                self.record = record
        return Renderer(record)

    @pytest.mark.usefixtures('load_plugin', 'lektor_context')
    def test(self, mixin):
        img = mixin.image('test.jpg', 'TITLE', 'ALT')
        assert re.match(r'<img.*>', img)
        assert 'alt="ALT"' in img
        assert 'srcset=' in img

    @pytest.mark.usefixtures('load_plugin', 'lektor_context')
    def test_not_resolvable(self, mixin):
        img = mixin.image('foo.jpg', 'TITLE', 'ALT')
        assert re.match(r'<img.*>', img)
        assert 'alt="ALT"' in img
        assert 'srcset=' not in img


class TestResponsiveImagePlugin(object):
    @pytest.fixture
    def plugin(self, lektor_env):
        return ResponsiveImagePlugin(lektor_env, 'responsive-image')

    def test_on_setup_env(self, plugin, lektor_env):
        plugin.on_setup_env()
        jinja_globals = lektor_env.jinja_env.globals
        assert 'responsive_image' in jinja_globals

    def test_on_markdown_config(self, plugin):
        config = MarkdownConfig()
        plugin.on_markdown_config(config)
        assert config.renderer_mixins == [ResponsiveImageMixin]

    def test_get_responsive_image_config(self, plugin):
        config = plugin.get_responsive_image_config()
        assert config == ResponsiveImage.DEFAULT_CONFIG
        assert config is not ResponsiveImage.DEFAULT_CONFIG

    @pytest.fixture
    def inifile(self, config, tmp_path):
        ini = IniFile(str(tmp_path / 'test.ini'))
        for key, value in config.items():
            ini['default.' + key] = value
        ini.save()
        return ini

    @pytest.mark.parametrize('config, expected', [
        ({'widths': '10 20 '}, {'widths': [10, 20]}),
        ({'widths': '10-20'}, {}),
        ({'quality': '12'}, {'quality': 12}),
        ({'quality': 'twelve'}, {}),
        ({'default_width': '10'}, {'default_width': 10}),
        ({'sizes': '60vw'}, {'sizes': '60vw'}),
        ])
    def test__parse_ini(self, plugin, inifile, expected):
        assert plugin._parse_ini(inifile) == expected



# -*- coding: utf-8 -*-
from functools import wraps
import inspect
try:
    from urllib.parse import urlparse
except ImportError:             # pragma: no cover
    from urlparse import urlparse

from lektor.context import get_ctx
from lektor.db import Image
from lektor.pluginsystem import (
    Plugin,
    get_plugin,
    )
from lektor.utils import join_path
from markupsafe import escape
from werkzeug.utils import cached_property


def ignore_unsupported_kwargs(f):
    getargspec = inspect.getfullargspec if hasattr(inspect, 'getfullargspec') \
                 else inspect.getargspec
    supported_args = getargspec(f).args

    @wraps(f)
    def wrapped(*args, **kwargs):
        supported_kwargs = set(supported_args[len(args):]).intersection(kwargs)
        return f(*args, **{k: kwargs[k] for k in supported_kwargs})
    return wrapped


def fmt_attrs(attrs):
    return ' '.join('{}="{}"'.format(key, escape(val))
                    for key, val in attrs.items()
                    if val is not None)


class ResponsiveImage(object):
    DEFAULT_CONFIG = {
        'widths': [480, 800, 1200, 2400],
        'quality': 92,
        'default_width': 1200,
        'sizes': None,
        }

    def __init__(self, image, config=None, ctx=None):
        if config is None:
            config = self.DEFAULT_CONFIG.copy()
        if ctx is None:
            ctx = get_ctx()
        self.config = config
        self.image = image
        self.ctx = ctx

    @property
    def attrs(self):
        attrs = self.default_image_attrs()
        if self.srcset is not None:
            attrs['srcset'] = self.srcset
            if self.sizes is not None:
                attrs['sizes'] = self.sizes
        return attrs

    @cached_property
    def srcset(self):
        url_to = self.ctx.url_to
        widths = list(self.iter_widths())
        if len(widths) > 1:
            def image_info(width):
                image = self.resize_image(width)
                return "{url} {width}w".format(url=url_to(image), width=width)
            return ', '.join(map(image_info, self.iter_widths()))

    @property
    def sizes(self):
        return self.config.get('sizes')

    def iter_widths(self):
        image_width = self.image.width
        widths = self.config.get('widths', [])
        for w in widths:
            if w < image_width:
                yield w
            else:
                yield image_width
                break

    def default_image_attrs(self):
        default_width = self.config.get('default_width', 800)
        default_image = self.resize_image(default_width)
        return {
            'src': self.ctx.url_to(default_image),
            'width': default_image.width,
            'height': default_image.height,
            }

    def resize_image(self, width):
        image = self.image
        if width is None or width >= image.width:
            return image
        # We (should) never upscale.  Upscale=True is passed here
        # solely to avoid triggering a deprecation warning.
        thumbnail = ignore_unsupported_kwargs(self.image.thumbnail)
        return thumbnail(width, quality=self.config['quality'], upscale=True)


def resolve_image(record, src):
    if record is None:
        return None
    url = urlparse(src)
    if url.scheme or url.netloc:
        return None

    path = url.path
    if not path.startswith('/'):
        assert record.path is not None   # FIXME: better error reporting?
        path = join_path(record.path, path)
    source = record.pad.get(path)
    if isinstance(source, Image):
        return source


class ResponsiveImageMixin(object):
    """Render markdown images at responsive resolutions (using srcset).

    """

    def image(self, src, title, text):
        image = resolve_image(self.record, src)
        if image is not None and image.format in ('png', 'gif', 'jpeg'):
            plugin = get_plugin('responsive-image', env=self.record.pad.env)
            attrs = {'alt': text, 'title': title or None}
            attrs.update(plugin.responsive_image(image).attrs)
            return "<img {}>".format(fmt_attrs(attrs))
        else:
            return super(ResponsiveImageMixin, self).image(src, title, text)


class ResponsiveImagePlugin(Plugin):
    name = 'Responsive Image'
    description = u'Support for responsive-resolutioned images.'

    def on_setup_env(self, **extra):
        self.env.jinja_env.globals['responsive_image'] = self.responsive_image

    def on_markdown_config(self, config, **extra):
        config.renderer_mixins.append(ResponsiveImageMixin)

    def responsive_image(self, image):
        config = self.get_responsive_image_config()
        return ResponsiveImage(image, config)

    def get_responsive_image_config(self):
        inifile = self.get_config()
        config = ResponsiveImage.DEFAULT_CONFIG.copy()
        config.update(self._parse_ini(inifile))
        return config

    @staticmethod
    def _parse_ini(inifile, section='default'):
        data = inifile.section_as_dict(section)
        config = {}
        if 'widths' in data:
            try:
                config['widths'] = list(map(int, data['widths'].split()))
            except ValueError:
                pass
        for key in ('quality', 'default_width'):
            if key in data:
                try:
                    config[key] = int(data[key])
                except ValueError:
                    pass
        if 'sizes' in data:
            config['sizes'] = data['sizes']
        return config
